@@ -135,7 +135,7 @@ static const struct grub_arg_option openbsd_opts[] =
     {"single", 's', 0, N_("Boot into single mode."), 0, 0},
     {"kdb", 'd', 0, N_("Enter in KDB on boot."), 0, 0},
     {"root", 'r', 0, N_("Set root device."), "wdXY", ARG_TYPE_STRING},
-    {"serial", 'h', GRUB_ARG_OPTION_OPTIONAL, 
+    {"serial", 'h', GRUB_ARG_OPTION_OPTIONAL,
      N_("Use serial console."),
      /* TRANSLATORS: "com" is static and not to be translated. It refers to
 	serial ports e.g. com1.
@@ -168,7 +168,7 @@ static const struct grub_arg_option netbsd_opts[] =
     {"debug", 'x', 0, N_("Boot with debug messages."), 0, 0},
     {"silent", 'z', 0, N_("Suppress normal output (warnings remain)."), 0, 0},
     {"root", 'r', 0, N_("Set root device."), N_("DEVICE"), ARG_TYPE_STRING},
-    {"serial", 'h', GRUB_ARG_OPTION_OPTIONAL, 
+    {"serial", 'h', GRUB_ARG_OPTION_OPTIONAL,
      N_("Use serial console."),
      /* TRANSLATORS: "com" is static and not to be translated. It refers to
 	serial ports e.g. com1.
@@ -231,7 +231,7 @@ grub_bsd_add_meta_ptr (grub_uint32_t type, void **ptr, grub_uint32_t len)
   newtag->next = NULL;
   *ptr = newtag->data;
 
-  if (kernel_type == KERNEL_TYPE_FREEBSD 
+  if (kernel_type == KERNEL_TYPE_FREEBSD
       && type == (FREEBSD_MODINFO_METADATA | FREEBSD_MODINFOMD_SMAP))
     {
       struct bsd_tag *p;
@@ -263,7 +263,7 @@ grub_err_t
 grub_bsd_add_meta (grub_uint32_t type, const void *data, grub_uint32_t len)
 {
   grub_err_t err;
-  void *ptr;
+  void *ptr = NULL;
 
   err = grub_bsd_add_meta_ptr (type, &ptr, len);
   if (err)
@@ -435,7 +435,7 @@ grub_freebsd_add_meta_module (const char *filename, const char *type,
     {
       grub_uint64_t addr64 = addr, size64 = size;
       if (grub_bsd_add_meta (FREEBSD_MODINFO_TYPE, type, grub_strlen (type) + 1)
-	  || grub_bsd_add_meta (FREEBSD_MODINFO_ADDR, &addr64, sizeof (addr64)) 
+	  || grub_bsd_add_meta (FREEBSD_MODINFO_ADDR, &addr64, sizeof (addr64))
 	  || grub_bsd_add_meta (FREEBSD_MODINFO_SIZE, &size64, sizeof (size64)))
 	return grub_errno;
     }
@@ -633,7 +633,7 @@ grub_freebsd_boot (void)
       err = grub_bsd_add_meta (FREEBSD_MODINFO_END, 0, 0);
       if (err)
 	return err;
-      
+
       tag_buf_len = 0;
       for (tag = tags; tag; tag = tag->next)
 	tag_buf_len = ALIGN_VAR (tag_buf_len
@@ -678,7 +678,7 @@ grub_freebsd_boot (void)
     {
       grub_uint8_t *p_tag = p;
       struct bsd_tag *tag;
-      
+
       for (tag = tags; tag; tag = tag->next)
 	{
 	  struct freebsd_tag_header *head
@@ -728,7 +728,7 @@ grub_freebsd_boot (void)
 
   if (is_64bit)
     {
-      struct grub_relocator64_state state;
+      struct grub_relocator64_state state = {0};
       grub_uint8_t *pagetable;
       grub_uint32_t *stack;
       grub_addr_t stack_target;
@@ -767,7 +767,7 @@ grub_freebsd_boot (void)
     }
   else
     {
-      struct grub_relocator32_state state;
+      struct grub_relocator32_state state = {0};
       grub_uint32_t *stack;
       grub_addr_t stack_target;
 
@@ -842,7 +842,7 @@ grub_openbsd_boot (void)
     if (regs.edx == 0x20494350)
       {
 	struct grub_openbsd_bootarg_pcibios pcibios;
-	
+
 	pcibios.characteristics = regs.eax & 0xff;
 	pcibios.revision = regs.ebx & 0xffff;
 	pcibios.pm_entry = regs.edi;
@@ -929,7 +929,7 @@ grub_netbsd_setup_video (void)
   struct grub_video_mode_info mode_info;
   void *framebuffer;
   const char *modevar;
-  struct grub_netbsd_btinfo_framebuf params;
+  struct grub_netbsd_btinfo_framebuf params = {0};
   grub_err_t err;
   grub_video_driver_id_t driv_id;
 
@@ -1340,6 +1340,7 @@ static grub_err_t
 grub_bsd_load_elf (grub_elf_t elf, const char *filename)
 {
   grub_err_t err;
+  grub_size_t sz;
 
   kern_end = 0;
   kern_start = ~0;
@@ -1370,8 +1371,11 @@ grub_bsd_load_elf (grub_elf_t elf, const char *filename)
 
       if (grub_errno)
 	return grub_errno;
-      err = grub_relocator_alloc_chunk_addr (relocator, &ch,
-					     kern_start, kern_end - kern_start);
+
+      if (grub_sub (kern_end, kern_start, &sz))
+	return grub_error (GRUB_ERR_OUT_OF_RANGE, "underflow detected while determining size of kernel for relocator");
+
+      err = grub_relocator_alloc_chunk_addr (relocator, &ch, kern_start, sz);
       if (err)
 	return err;
 
@@ -1431,8 +1435,10 @@ grub_bsd_load_elf (grub_elf_t elf, const char *filename)
       {
 	grub_relocator_chunk_t ch;
 
-	err = grub_relocator_alloc_chunk_addr (relocator, &ch, kern_start,
-					       kern_end - kern_start);
+	if (grub_sub (kern_end, kern_start, &sz))
+	  return grub_error (GRUB_ERR_OUT_OF_RANGE, "underflow detected while determining size of kernel for relocator");
+
+	err = grub_relocator_alloc_chunk_addr (relocator, &ch, kern_start, sz);
 	if (err)
 	  return err;
 	kern_chunk_src = get_virtual_current_address (ch);
@@ -1500,7 +1506,7 @@ grub_bsd_load (int argc, char *argv[])
 fail:
 
   if (grub_errno != GRUB_ERR_NONE)
-    grub_dl_unref (my_mod);	
+    grub_dl_unref (my_mod);
 
   return grub_errno;
 }
@@ -1537,9 +1543,7 @@ grub_cmd_freebsd (grub_extcmd_context_t ctxt, int argc, char *argv[])
 	  grub_file_t file;
 	  int len = is_64bit ? 8 : 4;
 
-	  err = grub_freebsd_add_meta_module (argv[0], is_64bit
-					      ? FREEBSD_MODTYPE_KERNEL64
-					      : FREEBSD_MODTYPE_KERNEL,
+	  err = grub_freebsd_add_meta_module (argv[0], FREEBSD_MODTYPE_KERNEL,
 					      argc - 1, argv + 1,
 					      kern_start,
 					      kern_end - kern_start);
@@ -1658,7 +1662,7 @@ grub_cmd_openbsd (grub_extcmd_context_t ctxt, int argc, char *argv[])
 			       "only com0-com3 are supported");
 	  if (*ptr == ',')
 	    {
-	      ptr++; 
+	      ptr++;
 	      speed = grub_strtoul (ptr, &ptr, 0);
 	      if (grub_errno)
 		return grub_errno;
@@ -1668,7 +1672,7 @@ grub_cmd_openbsd (grub_extcmd_context_t ctxt, int argc, char *argv[])
       serial.device = (GRUB_OPENBSD_COM_MAJOR << 8) | port;
       serial.speed = speed;
       serial.addr = grub_ns8250_hw_get_port (port);
-	  
+
       grub_bsd_add_meta (OPENBSD_BOOTARG_CONSOLE, &serial, sizeof (serial));
       bootflags |= OPENBSD_RB_SERCONS;
     }
@@ -1755,7 +1759,7 @@ grub_cmd_netbsd (grub_extcmd_context_t ctxt, int argc, char *argv[])
 	      if (grub_memcmp (ptr, "com", sizeof ("com") - 1) == 0)
 		{
 		  ptr += sizeof ("com") - 1;
-		  serial.addr 
+		  serial.addr
 		    = grub_ns8250_hw_get_port (grub_strtoul (ptr, &ptr, 0));
 		}
 	      else
@@ -1925,7 +1929,7 @@ grub_cmd_freebsd_module (grub_command_t cmd __attribute__ ((unused)),
 
   {
     grub_relocator_chunk_t ch;
-    err = grub_relocator_alloc_chunk_addr (relocator, &ch, kern_end, 
+    err = grub_relocator_alloc_chunk_addr (relocator, &ch, kern_end,
 					   file->size);
     if (err)
       goto fail;
@@ -1976,7 +1980,7 @@ grub_netbsd_module_load (char *filename, grub_uint32_t type)
 
   {
     grub_relocator_chunk_t ch;
-    err = grub_relocator_alloc_chunk_addr (relocator, &ch, kern_end, 
+    err = grub_relocator_alloc_chunk_addr (relocator, &ch, kern_end,
 					   file->size);
     if (err)
       goto fail;
