@@ -25,6 +25,7 @@
 #include <grub/net/netbuff.h>
 #include <grub/mm.h>
 #include <grub/priority_queue.h>
+#include <grub/safemath.h>
 #include <grub/time.h>
 
 struct iphdr {
@@ -38,7 +39,7 @@ struct iphdr {
   grub_uint16_t chksum;
   grub_uint32_t src;
   grub_uint32_t dest;
-} __attribute__ ((packed)) ;
+} GRUB_PACKED ;
 
 enum
 {
@@ -56,7 +57,7 @@ struct ip6hdr {
   grub_uint8_t ttl;
   ip6addr src;
   ip6addr dest;
-} __attribute__ ((packed)) ;
+} GRUB_PACKED ;
 
 static int
 cmp (const void *a__, const void *b__)
@@ -157,7 +158,7 @@ send_fragmented (struct grub_net_network_level_interface * inf,
       iph->service = 0;
       iph->len = grub_cpu_to_be16 (len + sizeof (struct iphdr));
       iph->ident = grub_cpu_to_be16 (id);
-      iph->frags = grub_cpu_to_be16 (off | (((grub_ssize_t) len 
+      iph->frags = grub_cpu_to_be16 (off | (((grub_ssize_t) len
 					     == nb->tail - nb->data)
 					    ? 0 : MORE_FRAGMENTS));
       iph->ttl = 0xff;
@@ -191,15 +192,18 @@ grub_net_send_ip4_packet (struct grub_net_network_level_interface *inf,
 			  grub_net_ip_protocol_t proto)
 {
   struct iphdr *iph;
+  grub_err_t err;
 
   COMPILE_TIME_ASSERT (GRUB_NET_OUR_IPV4_HEADER_SIZE == sizeof (*iph));
 
   if (nb->tail - nb->data + sizeof (struct iphdr) > inf->card->mtu)
     return send_fragmented (inf, target, nb, proto, *ll_target_addr);
 
-  grub_netbuff_push (nb, sizeof (*iph));
-  iph = (struct iphdr *) nb->data;
+  err = grub_netbuff_push (nb, sizeof (*iph));
+  if (err)
+    return err;
 
+  iph = (struct iphdr *) nb->data;
   iph->verhdrlen = ((4 << 4) | 5);
   iph->service = 0;
   iph->len = grub_cpu_to_be16 (nb->tail - nb->data);
@@ -225,12 +229,13 @@ handle_dgram (struct grub_net_buff *nb,
 	      grub_net_ip_protocol_t proto,
 	      const grub_net_network_level_address_t *source,
 	      const grub_net_network_level_address_t *dest,
+              grub_uint16_t *vlantag,
 	      grub_uint8_t ttl)
 {
   struct grub_net_network_level_interface *inf = NULL;
   grub_err_t err;
   int multicast = 0;
-  
+
   /* DHCP needs special treatment since we don't know IP yet.  */
   {
     struct udphdr *udph;
@@ -250,7 +255,7 @@ handle_dgram (struct grub_net_buff *nb,
 	    if (expected != chk)
 	      {
 		grub_dprintf ("net", "Invalid UDP checksum. "
-			      "Expected %x, got %x\n", 
+			      "Expected %x, got %x\n",
 			      grub_be_to_cpu16 (expected),
 			      grub_be_to_cpu16 (chk));
 		grub_netbuff_free (nb);
@@ -267,7 +272,7 @@ handle_dgram (struct grub_net_buff *nb,
 	  }
 
 	bootp = (const struct grub_net_bootp_packet *) nb->data;
-	
+
 	FOR_NET_NETWORK_LEVEL_INTERFACES (inf)
 	  if (inf->card == card
 	      && inf->address.type == GRUB_NET_NETWORK_LEVEL_PROTOCOL_DHCP_RECV
@@ -275,7 +280,7 @@ handle_dgram (struct grub_net_buff *nb,
 	      && grub_memcmp (inf->hwaddress.mac, &bootp->mac_addr,
 			      sizeof (inf->hwaddress.mac)) == 0)
 	    {
-	      grub_net_process_dhcp (nb, inf->card);
+	      grub_net_process_dhcp (nb, inf);
 	      grub_netbuff_free (nb);
 	      return GRUB_ERR_NONE;
 	    }
@@ -290,6 +295,15 @@ handle_dgram (struct grub_net_buff *nb,
 	&& grub_net_addr_cmp (&inf->address, dest) == 0
 	&& grub_net_hwaddr_cmp (&inf->hwaddress, hwaddress) == 0)
       break;
+
+    /* Verify vlantag id */
+    if (inf->card == card && inf->vlantag != *vlantag)
+      {
+        grub_dprintf ("net", "invalid vlantag! %x != %x\n",
+                      inf->vlantag, *vlantag);
+        break;
+      }
+
     /* Solicited node multicast.  */
     if (inf->card == card
 	&& inf->address.type == GRUB_NET_NETWORK_LEVEL_PROTOCOL_IPV6
@@ -301,7 +315,7 @@ handle_dgram (struct grub_net_buff *nb,
 	&& hwaddress->type == GRUB_NET_LINK_LEVEL_PROTOCOL_ETHERNET
 	&& hwaddress->mac[0] == 0x33 && hwaddress->mac[1] == 0x33
 	&& hwaddress->mac[2] == 0xff
-	&& hwaddress->mac[3] == ((grub_be_to_cpu64 (inf->address.ipv6[1]) 
+	&& hwaddress->mac[3] == ((grub_be_to_cpu64 (inf->address.ipv6[1])
 				  >> 16) & 0xff)
 	&& hwaddress->mac[4] == ((grub_be_to_cpu64 (inf->address.ipv6[1])
 				  >> 8) & 0xff)
@@ -312,7 +326,7 @@ handle_dgram (struct grub_net_buff *nb,
 	break;
       }
   }
- 
+
   if (!inf && !(dest->type == GRUB_NET_NETWORK_LEVEL_PROTOCOL_IPV6
 		&& dest->ipv6[0] == grub_be_to_cpu64_compile_time (0xff02ULL
 								   << 48)
@@ -360,7 +374,9 @@ static void
 free_old_fragments (void)
 {
   struct reassemble *rsm, **prev;
-  grub_uint64_t limit_time = grub_get_time_ms () - 90000;
+  grub_uint64_t limit_time = grub_get_time_ms ();
+
+  limit_time = (limit_time > 90000) ? limit_time - 90000 : 0;
 
   for (prev = &reassembles, rsm = *prev; rsm; rsm = *prev)
     if (rsm->last_time < limit_time)
@@ -378,7 +394,8 @@ static grub_err_t
 grub_net_recv_ip4_packets (struct grub_net_buff *nb,
 			   struct grub_net_card *card,
 			   const grub_net_link_level_address_t *hwaddress,
-			   const grub_net_link_level_address_t *src_hwaddress)
+			   const grub_net_link_level_address_t *src_hwaddress,
+                           grub_uint16_t *vlantag)
 {
   struct iphdr *iph = (struct iphdr *) nb->data;
   grub_err_t err;
@@ -403,7 +420,7 @@ grub_net_recv_ip4_packets (struct grub_net_buff *nb,
 					    * sizeof (grub_uint32_t)))
     {
       grub_dprintf ("net", "IP packet too short: %" PRIdGRUB_SSIZE "\n",
-		    (nb->tail - nb->data));
+		    (grub_ssize_t) (nb->tail - nb->data));
       grub_netbuff_free (nb);
       return GRUB_ERR_NONE;
     }
@@ -423,7 +440,7 @@ grub_net_recv_ip4_packets (struct grub_net_buff *nb,
       }
     if (actual_size < expected_size)
       {
-	grub_dprintf ("net", "Cut IP packet actual: %" PRIuGRUB_SIZE 
+	grub_dprintf ("net", "Cut IP packet actual: %" PRIuGRUB_SIZE
 		      ", expected %" PRIuGRUB_SIZE "\n", actual_size,
 		      expected_size);
 	grub_netbuff_free (nb);
@@ -453,7 +470,7 @@ grub_net_recv_ip4_packets (struct grub_net_buff *nb,
       dest.ipv4 = iph->dest;
 
       return handle_dgram (nb, card, src_hwaddress, hwaddress, iph->protocol,
-			   &source, &dest, iph->ttl);
+			   &source, &dest, vlantag, iph->ttl);
     }
 
   for (prev = &reassembles, rsm = *prev; rsm; prev = &rsm->next, rsm = *prev)
@@ -496,7 +513,14 @@ grub_net_recv_ip4_packets (struct grub_net_buff *nb,
     {
       rsm->total_len = (8 * (grub_be_to_cpu16 (iph->frags) & OFFSET_MASK)
 			+ (nb->tail - nb->data));
-      rsm->total_len -= ((iph->verhdrlen & 0xf) * sizeof (grub_uint32_t));
+
+      if (grub_sub (rsm->total_len, (iph->verhdrlen & 0xf) * sizeof (grub_uint32_t),
+		    &rsm->total_len))
+	{
+	  grub_dprintf ("net", "IP reassembly size underflow\n");
+	  return GRUB_ERR_NONE;
+	}
+
       rsm->asm_netbuff = grub_netbuff_alloc (rsm->total_len);
       if (!rsm->asm_netbuff)
 	{
@@ -589,7 +613,7 @@ grub_net_recv_ip4_packets (struct grub_net_buff *nb,
       dest.ipv4 = dst;
 
       return handle_dgram (ret, card, src_hwaddress,
-			   hwaddress, proto, &source, &dest,
+			   hwaddress, proto, &source, &dest, vlantag,
 			   ttl);
     }
 }
@@ -602,16 +626,19 @@ grub_net_send_ip6_packet (struct grub_net_network_level_interface *inf,
 			  grub_net_ip_protocol_t proto)
 {
   struct ip6hdr *iph;
+  grub_err_t err;
 
   COMPILE_TIME_ASSERT (GRUB_NET_OUR_IPV6_HEADER_SIZE == sizeof (*iph));
 
   if (nb->tail - nb->data + sizeof (struct iphdr) > inf->card->mtu)
     return grub_error (GRUB_ERR_NET_PACKET_TOO_BIG, "packet too big");
 
-  grub_netbuff_push (nb, sizeof (*iph));
-  iph = (struct ip6hdr *) nb->data;
+  err = grub_netbuff_push (nb, sizeof (*iph));
+  if (err)
+    return err;
 
-  iph->version_class_flow = grub_cpu_to_be32 ((6 << 28));
+  iph = (struct ip6hdr *) nb->data;
+  iph->version_class_flow = grub_cpu_to_be32_compile_time ((6 << 28));
   iph->len = grub_cpu_to_be16 (nb->tail - nb->data - sizeof (*iph));
   iph->protocol = proto;
   iph->ttl = 0xff;
@@ -644,7 +671,8 @@ static grub_err_t
 grub_net_recv_ip6_packets (struct grub_net_buff *nb,
 			   struct grub_net_card *card,
 			   const grub_net_link_level_address_t *hwaddress,
-			   const grub_net_link_level_address_t *src_hwaddress)
+			   const grub_net_link_level_address_t *src_hwaddress,
+                           grub_uint16_t *vlantag)
 {
   struct ip6hdr *iph = (struct ip6hdr *) nb->data;
   grub_err_t err;
@@ -654,7 +682,7 @@ grub_net_recv_ip6_packets (struct grub_net_buff *nb,
   if (nb->tail - nb->data < (grub_ssize_t) sizeof (*iph))
     {
       grub_dprintf ("net", "IP packet too short: %" PRIdGRUB_SSIZE "\n",
-		    nb->tail - nb->data);
+		    (grub_ssize_t) (nb->tail - nb->data));
       grub_netbuff_free (nb);
       return GRUB_ERR_NONE;
     }
@@ -681,7 +709,7 @@ grub_net_recv_ip6_packets (struct grub_net_buff *nb,
       }
     if (actual_size < expected_size)
       {
-	grub_dprintf ("net", "Cut IP packet actual: %" PRIuGRUB_SIZE 
+	grub_dprintf ("net", "Cut IP packet actual: %" PRIuGRUB_SIZE
 		      ", expected %" PRIuGRUB_SIZE "\n", actual_size,
 		      expected_size);
 	grub_netbuff_free (nb);
@@ -695,21 +723,24 @@ grub_net_recv_ip6_packets (struct grub_net_buff *nb,
   grub_memcpy (dest.ipv6, &iph->dest, sizeof (dest.ipv6));
 
   return handle_dgram (nb, card, src_hwaddress, hwaddress, iph->protocol,
-		       &source, &dest, iph->ttl);
+		       &source, &dest, vlantag, iph->ttl);
 }
 
 grub_err_t
 grub_net_recv_ip_packets (struct grub_net_buff *nb,
 			  struct grub_net_card *card,
 			  const grub_net_link_level_address_t *hwaddress,
-			  const grub_net_link_level_address_t *src_hwaddress)
+			  const grub_net_link_level_address_t *src_hwaddress,
+                          grub_uint16_t *vlantag)
 {
   struct iphdr *iph = (struct iphdr *) nb->data;
 
   if ((iph->verhdrlen >> 4) == 4)
-    return grub_net_recv_ip4_packets (nb, card, hwaddress, src_hwaddress);
+    return grub_net_recv_ip4_packets (nb, card, hwaddress, src_hwaddress,
+                                      vlantag);
   if ((iph->verhdrlen >> 4) == 6)
-    return grub_net_recv_ip6_packets (nb, card, hwaddress, src_hwaddress);
+    return grub_net_recv_ip6_packets (nb, card, hwaddress, src_hwaddress,
+                                      vlantag);
   grub_dprintf ("net", "Bad IP version: %d\n", (iph->verhdrlen >> 4));
   grub_netbuff_free (nb);
   return GRUB_ERR_NONE;
