@@ -35,12 +35,7 @@
 #include <limits.h>
 #endif
 
-#if defined(MAJOR_IN_MKDEV)
-#include <sys/mkdev.h>
-#elif defined(MAJOR_IN_SYSMACROS)
-#include <sys/sysmacros.h>
-#endif
-
+#include <grub/osdep/major.h>
 #include <grub/types.h>
 #include <sys/ioctl.h>         /* ioctl */
 #include <sys/mount.h>
@@ -49,6 +44,7 @@
 
 #include <grub/mm.h>
 #include <grub/misc.h>
+#include <grub/disk.h>
 #include <grub/emu/misc.h>
 #include <grub/emu/hostdisk.h>
 #include <grub/emu/getroot.h>
@@ -123,7 +119,7 @@ struct btrfs_ioctl_search_args {
                                struct btrfs_ioctl_fs_info_args)
 
 static int
-grub_util_is_imsm (const char *os_dev);
+grub_util_is_imsm_or_ddf (const char *os_dev);
 
 
 #define ESCAPED_PATH_MAX (4 * PATH_MAX)
@@ -135,10 +131,12 @@ struct mountinfo_entry
   char fstype[ESCAPED_PATH_MAX + 1], device[ESCAPED_PATH_MAX + 1];
 };
 
+#ifdef GRUB_UTIL
 static char **
 grub_util_raid_getmembers (const char *name, int bootable)
 {
   int fd, ret, i, j;
+  int remaining;
   char **devicelist;
   mdu_version_t version;
   mdu_array_info_t info;
@@ -170,22 +168,22 @@ grub_util_raid_getmembers (const char *name, int bootable)
 
   devicelist = xcalloc (info.nr_disks + 1, sizeof (char *));
 
-  for (i = 0, j = 0; j < info.nr_disks; i++)
+  remaining = info.nr_disks;
+  for (i = 0, j = 0; i < GRUB_MDRAID_MAX_DISKS && remaining > 0; i++)
     {
       disk.number = i;
       ret = ioctl (fd, GET_DISK_INFO, &disk);
       if (ret != 0)
 	grub_util_error (_("ioctl GET_DISK_INFO error: %s"), strerror (errno));
-      
+
+      /* Skip: MD_DISK_REMOVED slots don't contribute to "remaining" count. */
       if (disk.state & (1 << MD_DISK_REMOVED))
 	continue;
+      remaining--;
 
+      /* Only record disks that are actively participating in the array. */
       if (disk.state & (1 << MD_DISK_ACTIVE))
-	devicelist[j] = grub_find_device (NULL,
-					  makedev (disk.major, disk.minor));
-      else
-	devicelist[j] = NULL;
-      j++;
+        devicelist[j++] = grub_find_device (NULL, makedev (disk.major, disk.minor));
     }
 
   devicelist[j] = NULL;
@@ -194,6 +192,7 @@ grub_util_raid_getmembers (const char *name, int bootable)
 
   return devicelist;
 }
+#endif
 
 /* Statting something on a btrfs filesystem always returns a virtual device
    major/minor pair rather than the real underlying device, because btrfs
@@ -273,12 +272,12 @@ get_btrfs_fs_prefix (const char *mount_path)
   char *ret = NULL;
 
   fd = open (mount_path, O_RDONLY);
-	  
+
   if (fd < 0)
     return NULL;
   memset (&args, 0, sizeof(args));
   args.objectid = GRUB_BTRFS_TREE_ROOT_OBJECTID;
-  
+
   if (ioctl (fd, BTRFS_IOC_INO_LOOKUP, &args) < 0)
     goto fail;
   tree_id = args.treeid;
@@ -487,6 +486,9 @@ again:
 	}
     }
 
+  if (!entry_len)
+    goto out;
+
   /* Now scan visible mounts for the ones we're interested in.  */
   for (i = entry_len - 1; i >= 0; i--)
     {
@@ -579,6 +581,7 @@ out:
   return ret;
 }
 
+#ifdef GRUB_UTIL
 static char *
 get_mdadm_uuid (const char *os_dev)
 {
@@ -615,7 +618,7 @@ get_mdadm_uuid (const char *os_dev)
       if (strncmp (buf, "MD_UUID=", sizeof ("MD_UUID=") - 1) == 0)
 	{
 	  char *name_start, *ptri, *ptro;
-	  
+
 	  free (name);
 	  name_start = buf + sizeof ("MD_UUID=") - 1;
 	  ptro = name = xmalloc (strlen (name_start) + 1);
@@ -636,12 +639,13 @@ out:
 
   return name;
 }
+#endif
 
 static int
-grub_util_is_imsm (const char *os_dev)
+grub_util_is_imsm_or_ddf (const char *os_dev)
 {
   int retry;
-  int is_imsm = 0;
+  int is_imsm_or_ddf = 0;
   int container_seen = 0;
   const char *dev = os_dev;
 
@@ -702,8 +706,15 @@ grub_util_is_imsm (const char *os_dev)
 	  if (strncmp (buf, "MD_METADATA=imsm",
 		       sizeof ("MD_METADATA=imsm") - 1) == 0)
 	    {
-	      is_imsm = 1;
-	      grub_util_info ("%s is imsm", dev);	      
+	      is_imsm_or_ddf = 1;
+	      grub_util_info ("%s is imsm", dev);
+	      break;
+	    }
+	  if (strncmp (buf, "MD_METADATA=ddf",
+		       sizeof ("MD_METADATA=ddf") - 1) == 0)
+	    {
+	      is_imsm_or_ddf = 1;
+	      grub_util_info ("%s is ddf", dev);
 	      break;
 	    }
 	}
@@ -716,7 +727,7 @@ grub_util_is_imsm (const char *os_dev)
 
   if (dev != os_dev)
     free ((void *) dev);
-  return is_imsm;
+  return is_imsm_or_ddf;
 }
 
 char *
@@ -968,6 +979,7 @@ grub_util_part_to_disk (const char *os_dev, struct stat *st,
   return path;
 }
 
+#ifdef GRUB_UTIL
 static char *
 grub_util_get_raid_grub_dev (const char *os_dev)
 {
@@ -979,6 +991,8 @@ grub_util_get_raid_grub_dev (const char *os_dev)
       char *p, *q;
 
       p = strdup (os_dev + sizeof ("/dev/md_d") - 1);
+      if (p == NULL)
+        return NULL;
 
       q = strchr (p, 'p');
       if (q)
@@ -994,6 +1008,8 @@ grub_util_get_raid_grub_dev (const char *os_dev)
       char *p, *q;
 
       p = strdup (os_dev + sizeof ("/dev/md/d") - 1);
+      if (p == NULL)
+        return NULL;
 
       q = strchr (p, 'p');
       if (q)
@@ -1007,6 +1023,8 @@ grub_util_get_raid_grub_dev (const char *os_dev)
       char *p , *q;
 
       p = strdup (os_dev + sizeof ("/dev/md") - 1);
+      if (p == NULL)
+        return NULL;
 
       q = strchr (p, 'p');
       if (q)
@@ -1020,6 +1038,8 @@ grub_util_get_raid_grub_dev (const char *os_dev)
       char *p , *q;
 
       p = strdup (os_dev + sizeof ("/dev/md/") - 1);
+      if (p == NULL)
+        return NULL;
 
       q = strchr (p, 'p');
       if (q)
@@ -1034,6 +1054,8 @@ grub_util_get_raid_grub_dev (const char *os_dev)
       char *p , *q;
 
       p = strdup (os_dev + sizeof ("/dev/md/") - 1);
+      if (p == NULL)
+        return NULL;
 
       q = strchr (p, 'p');
       if (q)
@@ -1070,6 +1092,7 @@ grub_util_get_raid_grub_dev (const char *os_dev)
   }
   return grub_dev;
 }
+#endif
 
 enum grub_dev_abstraction_types
 grub_util_get_dev_abstraction_os (const char *os_dev)
@@ -1081,11 +1104,12 @@ grub_util_get_dev_abstraction_os (const char *os_dev)
 
   /* Check for RAID.  */
   if (!strncmp (os_dev, "/dev/md", 7) && ! grub_util_device_is_mapped (os_dev)
-      && !grub_util_is_imsm (os_dev))
+      && !grub_util_is_imsm_or_ddf (os_dev))
     return GRUB_DEV_ABSTRACTION_RAID;
   return GRUB_DEV_ABSTRACTION_NONE;
 }
 
+#ifdef GRUB_UTIL
 int
 grub_util_pull_device_os (const char *os_dev,
 			  enum grub_dev_abstraction_types ab)
@@ -1142,6 +1166,7 @@ grub_util_get_grub_dev_os (const char *os_dev)
 
   return grub_dev;
 }
+#endif
 
 char *
 grub_make_system_path_relative_to_its_root_os (const char *path)
