@@ -21,7 +21,7 @@
 
 typedef unsigned long grub_jmp_buf[10];
 
-int grub_setjmp (grub_jmp_buf env) __attribute__ ((returns_twice));
+int grub_setjmp (grub_jmp_buf env) RETURNS_TWICE;
 void grub_longjmp (grub_jmp_buf env, int val) __attribute__ ((noreturn));
 
 #endif /* ! GRUB_SETJMP_CPU_HEADER */
