@@ -23,6 +23,7 @@
 #include <grub/device.h>
 #include <grub/symbol.h>
 #include <grub/types.h>
+#include <grub/dl.h>
 
 #include <grub/list.h>
 /* For embedding types.  */
@@ -38,7 +39,9 @@ struct grub_dirhook_info
   unsigned dir:1;
   unsigned mtimeset:1;
   unsigned case_insensitive:1;
-  grub_int32_t mtime;
+  unsigned inodeset:1;
+  grub_int64_t mtime;
+  grub_uint64_t inode;
 };
 
 typedef int (*grub_fs_dir_hook_t) (const char *filename,
@@ -55,35 +58,38 @@ struct grub_fs
   /* My name.  */
   const char *name;
 
+  /* My module */
+  grub_dl_t mod;
+
   /* Call HOOK with each file under DIR.  */
-  grub_err_t (*dir) (grub_device_t device, const char *path,
+  grub_err_t (*fs_dir) (grub_device_t device, const char *path,
 		     grub_fs_dir_hook_t hook, void *hook_data);
 
   /* Open a file named NAME and initialize FILE.  */
-  grub_err_t (*open) (struct grub_file *file, const char *name);
+  grub_err_t (*fs_open) (struct grub_file *file, const char *name);
 
   /* Read LEN bytes data from FILE into BUF.  */
-  grub_ssize_t (*read) (struct grub_file *file, char *buf, grub_size_t len);
+  grub_ssize_t (*fs_read) (struct grub_file *file, char *buf, grub_size_t len);
 
   /* Close the file FILE.  */
-  grub_err_t (*close) (struct grub_file *file);
+  grub_err_t (*fs_close) (struct grub_file *file);
 
   /* Return the label of the device DEVICE in LABEL.  The label is
      returned in a grub_malloc'ed buffer and should be freed by the
      caller.  */
-  grub_err_t (*label) (grub_device_t device, char **label);
+  grub_err_t (*fs_label) (grub_device_t device, char **label);
 
   /* Return the uuid of the device DEVICE in UUID.  The uuid is
      returned in a grub_malloc'ed buffer and should be freed by the
      caller.  */
-  grub_err_t (*uuid) (grub_device_t device, char **uuid);
+  grub_err_t (*fs_uuid) (grub_device_t device, char **uuid);
 
   /* Get writing time of filesystem. */
-  grub_err_t (*mtime) (grub_device_t device, grub_int32_t *timebuf);
+  grub_err_t (*fs_mtime) (grub_device_t device, grub_int64_t *timebuf);
 
 #ifdef GRUB_UTIL
   /* Determine sectors available for embedding.  */
-  grub_err_t (*embed) (grub_device_t device, unsigned int *nsectors,
+  grub_err_t (*fs_embed) (grub_device_t device, unsigned int *nsectors,
 		       unsigned int max_nsectors,
 		       grub_embed_type_t embed_type,
 		       grub_disk_addr_t **sectors);
@@ -126,4 +132,6 @@ grub_fs_unregister (grub_fs_t fs)
 
 grub_fs_t EXPORT_FUNC(grub_fs_probe) (grub_device_t device);
 
+#define GRUB_ENV_BTRFS_OFFSET (256 * 1024)
+
 #endif /* ! GRUB_FS_HEADER */
