@@ -22,6 +22,10 @@
 #include <grub/misc.h>
 #include <grub/err.h>
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 struct grub_priority_queue;
 typedef struct grub_priority_queue *grub_priority_queue_t;
 typedef int (*grub_comparator_t) (const void *a, const void *b);
@@ -33,4 +37,8 @@ void *grub_priority_queue_top (grub_priority_queue_t pq);
 void grub_priority_queue_pop (grub_priority_queue_t pq);
 grub_err_t grub_priority_queue_push (grub_priority_queue_t pq, const void *el);
 
+#ifdef __cplusplus
+}
+#endif
+
 #endif
